"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonPatchDraft = void 0;
const ArrayType_1 = require("../types/rga-array/ArrayType");
const deepEqual_1 = require("../../json-equal/deepEqual");
const Draft_1 = require("../../json-crdt-patch/Draft");
const json_pointer_1 = require("../../json-pointer");
const ObjectType_1 = require("../types/lww-object/ObjectType");
const constants_1 = require("../../json-crdt-patch/constants");
const util_1 = require("../../json-pointer/util");
class JsonPatchDraft {
    constructor(model) {
        this.model = model;
        this.draft = new Draft_1.Draft();
    }
    applyOps(ops) {
        for (const op of ops)
            this.applyOp(op);
    }
    applyOp(op) {
        switch (op.op) {
            case 'add':
                this.applyOpAdd(op);
                break;
            case 'remove':
                this.applyRemove(op);
                break;
            case 'replace':
                this.applyReplace(op);
                break;
            case 'move':
                this.applyMove(op);
                break;
            case 'copy':
                this.applyCopy(op);
                break;
            case 'test':
                this.applyTest(op);
                break;
            case 'str_ins':
                this.applyStrIns(op);
                break;
            case 'str_del':
                this.applyStrDel(op);
                break;
            default:
                throw new Error('UNKNOWN_OP');
        }
    }
    applyOpAdd(op) {
        const { builder } = this.draft;
        const steps = (0, util_1.toPath)(op.path);
        if (!steps.length)
            this.setRoot(op.value);
        else {
            const objSteps = steps.slice(0, steps.length - 1);
            const node = this.model.api.find(objSteps);
            const key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                builder.setKeys(node.id, [[String(key), builder.json(op.value)]]);
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const value = builder.json(op.value);
                if (key === '-') {
                    const length = node.size();
                    const after = node.findId(length);
                    builder.insArr(node.id, after, [value]);
                }
                else {
                    const index = ~~key;
                    if ('' + index !== key)
                        throw new Error('INVALID_INDEX');
                    if (!index)
                        builder.insArr(node.id, node.id, [value]);
                    else {
                        const after = node.findId(index - 1);
                        builder.insArr(node.id, after, [value]);
                    }
                }
            }
            else
                throw new Error('NOT_FOUND');
        }
    }
    applyRemove(op) {
        const { builder } = this.draft;
        const steps = (0, util_1.toPath)(op.path);
        if (!steps.length)
            this.setRoot(null);
        else {
            const objSteps = steps.slice(0, steps.length - 1);
            const node = this.model.api.find(objSteps);
            const key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                const stringKey = String(key);
                if (node.get(stringKey) === undefined)
                    throw new Error('NOT_FOUND');
                builder.setKeys(node.id, [[stringKey, constants_1.UNDEFINED_ID]]);
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const key = steps[steps.length - 1];
                const index = ~~key;
                if ('' + index !== key)
                    throw new Error('INVALID_INDEX');
                const id = node.findId(index);
                builder.del(node.id, id, 1);
            }
            else
                throw new Error('NOT_FOUND');
        }
    }
    applyReplace(op) {
        const { path, value } = op;
        this.applyRemove({ op: 'remove', path });
        this.applyOpAdd({ op: 'add', path, value });
    }
    applyMove(op) {
        const path = (0, util_1.toPath)(op.path);
        const from = (0, util_1.toPath)(op.from);
        if ((0, json_pointer_1.isChild)(from, path))
            throw new Error('INVALID_CHILD');
        const json = this.json(from);
        this.applyRemove({ op: 'remove', path: from });
        this.applyOpAdd({ op: 'add', path, value: json });
    }
    applyCopy(op) {
        const path = (0, util_1.toPath)(op.path);
        const from = (0, util_1.toPath)(op.from);
        const json = this.json(from);
        this.applyOpAdd({ op: 'add', path, value: json });
    }
    applyTest(op) {
        const path = (0, util_1.toPath)(op.path);
        const json = this.json(path);
        if (!(0, deepEqual_1.deepEqual)(json, op.value))
            throw new Error('TEST');
    }
    applyStrIns(op) {
        const path = (0, util_1.toPath)(op.path);
        const { node } = this.model.api.str(path);
        const { builder } = this.draft;
        const length = node.length();
        const after = op.pos ? node.findId(length < op.pos ? length - 1 : op.pos - 1) : node.id;
        builder.insStr(node.id, after, op.str);
    }
    applyStrDel(op) {
        const path = (0, util_1.toPath)(op.path);
        const { node } = this.model.api.str(path);
        const { builder } = this.draft;
        const length = node.length();
        if (length <= op.pos)
            return;
        const after = node.findId(op.pos);
        const deletionLength = Math.min(op.len ?? op.str.length, length - op.pos);
        builder.del(node.id, after, deletionLength);
    }
    get(steps) {
        if (!steps.length)
            return this.model.toView();
        else {
            const objSteps = steps.slice(0, steps.length - 1);
            const node = this.model.api.find(objSteps);
            const key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                return node.getNode(String(key))?.toJson();
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const index = ~~key;
                if ('' + index !== key)
                    throw new Error('INVALID_INDEX');
                const nodeId = node.findValue(index);
                return this.model.node(nodeId)?.toJson();
            }
        }
        return undefined;
    }
    json(steps) {
        const json = this.get(steps);
        if (json === undefined)
            throw new Error('NOT_FOUND');
        return json;
    }
    setRoot(json) {
        const { builder } = this.draft;
        builder.root(builder.json(json));
    }
}
exports.JsonPatchDraft = JsonPatchDraft;
