"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonPatchDraft = void 0;
var tslib_1 = require("tslib");
var ArrayType_1 = require("../types/rga-array/ArrayType");
var deepEqual_1 = require("../../json-equal/deepEqual");
var Draft_1 = require("../../json-crdt-patch/Draft");
var json_pointer_1 = require("../../json-pointer");
var ObjectType_1 = require("../types/lww-object/ObjectType");
var constants_1 = require("../../json-crdt-patch/constants");
var util_1 = require("../../json-pointer/util");
var JsonPatchDraft = (function () {
    function JsonPatchDraft(model) {
        this.model = model;
        this.draft = new Draft_1.Draft();
    }
    JsonPatchDraft.prototype.applyOps = function (ops) {
        var e_1, _a;
        try {
            for (var ops_1 = (0, tslib_1.__values)(ops), ops_1_1 = ops_1.next(); !ops_1_1.done; ops_1_1 = ops_1.next()) {
                var op = ops_1_1.value;
                this.applyOp(op);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (ops_1_1 && !ops_1_1.done && (_a = ops_1.return)) _a.call(ops_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    };
    JsonPatchDraft.prototype.applyOp = function (op) {
        switch (op.op) {
            case 'add':
                this.applyOpAdd(op);
                break;
            case 'remove':
                this.applyRemove(op);
                break;
            case 'replace':
                this.applyReplace(op);
                break;
            case 'move':
                this.applyMove(op);
                break;
            case 'copy':
                this.applyCopy(op);
                break;
            case 'test':
                this.applyTest(op);
                break;
            case 'str_ins':
                this.applyStrIns(op);
                break;
            case 'str_del':
                this.applyStrDel(op);
                break;
            default:
                throw new Error('UNKNOWN_OP');
        }
    };
    JsonPatchDraft.prototype.applyOpAdd = function (op) {
        var builder = this.draft.builder;
        var steps = (0, util_1.toPath)(op.path);
        if (!steps.length)
            this.setRoot(op.value);
        else {
            var objSteps = steps.slice(0, steps.length - 1);
            var node = this.model.api.find(objSteps);
            var key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                builder.setKeys(node.id, [[String(key), builder.json(op.value)]]);
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                var value = builder.json(op.value);
                if (key === '-') {
                    var length_1 = node.size();
                    var after = node.findId(length_1);
                    builder.insArr(node.id, after, [value]);
                }
                else {
                    var index = ~~key;
                    if ('' + index !== key)
                        throw new Error('INVALID_INDEX');
                    if (!index)
                        builder.insArr(node.id, node.id, [value]);
                    else {
                        var after = node.findId(index - 1);
                        builder.insArr(node.id, after, [value]);
                    }
                }
            }
            else
                throw new Error('NOT_FOUND');
        }
    };
    JsonPatchDraft.prototype.applyRemove = function (op) {
        var builder = this.draft.builder;
        var steps = (0, util_1.toPath)(op.path);
        if (!steps.length)
            this.setRoot(null);
        else {
            var objSteps = steps.slice(0, steps.length - 1);
            var node = this.model.api.find(objSteps);
            var key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                var stringKey = String(key);
                if (node.get(stringKey) === undefined)
                    throw new Error('NOT_FOUND');
                builder.setKeys(node.id, [[stringKey, constants_1.UNDEFINED_ID]]);
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                var key_1 = steps[steps.length - 1];
                var index = ~~key_1;
                if ('' + index !== key_1)
                    throw new Error('INVALID_INDEX');
                var id = node.findId(index);
                builder.del(node.id, id, 1);
            }
            else
                throw new Error('NOT_FOUND');
        }
    };
    JsonPatchDraft.prototype.applyReplace = function (op) {
        var path = op.path, value = op.value;
        this.applyRemove({ op: 'remove', path: path });
        this.applyOpAdd({ op: 'add', path: path, value: value });
    };
    JsonPatchDraft.prototype.applyMove = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var from = (0, util_1.toPath)(op.from);
        if ((0, json_pointer_1.isChild)(from, path))
            throw new Error('INVALID_CHILD');
        var json = this.json(from);
        this.applyRemove({ op: 'remove', path: from });
        this.applyOpAdd({ op: 'add', path: path, value: json });
    };
    JsonPatchDraft.prototype.applyCopy = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var from = (0, util_1.toPath)(op.from);
        var json = this.json(from);
        this.applyOpAdd({ op: 'add', path: path, value: json });
    };
    JsonPatchDraft.prototype.applyTest = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var json = this.json(path);
        if (!(0, deepEqual_1.deepEqual)(json, op.value))
            throw new Error('TEST');
    };
    JsonPatchDraft.prototype.applyStrIns = function (op) {
        var path = (0, util_1.toPath)(op.path);
        var node = this.model.api.str(path).node;
        var builder = this.draft.builder;
        var length = node.length();
        var after = op.pos ? node.findId(length < op.pos ? length - 1 : op.pos - 1) : node.id;
        builder.insStr(node.id, after, op.str);
    };
    JsonPatchDraft.prototype.applyStrDel = function (op) {
        var _a;
        var path = (0, util_1.toPath)(op.path);
        var node = this.model.api.str(path).node;
        var builder = this.draft.builder;
        var length = node.length();
        if (length <= op.pos)
            return;
        var after = node.findId(op.pos);
        var deletionLength = Math.min((_a = op.len) !== null && _a !== void 0 ? _a : op.str.length, length - op.pos);
        builder.del(node.id, after, deletionLength);
    };
    JsonPatchDraft.prototype.get = function (steps) {
        var _a, _b;
        if (!steps.length)
            return this.model.toView();
        else {
            var objSteps = steps.slice(0, steps.length - 1);
            var node = this.model.api.find(objSteps);
            var key = steps[steps.length - 1];
            if (node instanceof ObjectType_1.ObjectType) {
                return (_a = node.getNode(String(key))) === null || _a === void 0 ? void 0 : _a.toJson();
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                var index = ~~key;
                if ('' + index !== key)
                    throw new Error('INVALID_INDEX');
                var nodeId = node.findValue(index);
                return (_b = this.model.node(nodeId)) === null || _b === void 0 ? void 0 : _b.toJson();
            }
        }
        return undefined;
    };
    JsonPatchDraft.prototype.json = function (steps) {
        var json = this.get(steps);
        if (json === undefined)
            throw new Error('NOT_FOUND');
        return json;
    };
    JsonPatchDraft.prototype.setRoot = function (json) {
        var builder = this.draft.builder;
        builder.root(builder.json(json));
    };
    return JsonPatchDraft;
}());
exports.JsonPatchDraft = JsonPatchDraft;
